package com.untzuntz.ustackclientapi;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

public class ParameterSet {

	private HashMap<ParameterDefinition,String> params = new HashMap<ParameterDefinition,String>();
	public static final DateFormat df = new SimpleDateFormat("yyyyMMddHHmmss");
	
	public ParameterSet(ParameterDefinition k, String v)
	{
		add(k, v);
	}
	
	public ParameterSet(ParameterDefinition k, Date v)
	{
		if (v != null)
			add(k, df.format(v));
	}
	
	public ParameterSet(ParameterDefinition k, Long v)
	{
		add(k, v);
	}
	
	public ParameterSet(ParameterDefinition k, Integer v)
	{
		add(k, v);
	}
	
	public ParameterSet(ParameterDefinition k, Object v)
	{
		add(k, v);
	}
	
	public ParameterSet add(ParameterDefinition k, Long v)
	{
		if (k == null || v == null)
			return this;
		add(k, v.toString());
		return this;
	}

	public ParameterSet add(ParameterDefinition k, Integer v)
	{
		if (k == null || v == null)
			return this;
		add(k, v.toString());
		return this;
	}
	
	public ParameterSet add(ParameterDefinition k, Rectangle v)
	{
		if (k == null || v == null)
			return this;
		
		add(k, String.format("%dx%d", v.getWidth(), v.getHeight()));
		return this;
	}

	public ParameterSet add(ParameterDefinition k, Point v)
	{
		if (k == null || v == null)
			return this;
		
		add(k, String.format("%dx%d", v.x, v.y));
		return this;
	}

	public ParameterSet add(ParameterDefinition k, Date v)
	{
		if (v != null)
			add(k, df.format(v));
		
		return this;
	}

	public ParameterSet add(ParameterDefinition k, Object v)
	{
		if (k == null || v == null)
			return this;
		
		add(k, v.toString());
		return this;
	}
	
	public ParameterSet add(ParameterDefinition k, String v)
	{
		if (v != null && k != null)
			params.put(k, v);
		return this;
	}

	public ParameterSet add(ParameterDefinition k, String[] v)
	{
		if (v != null && k != null)
		{
			StringBuffer buf = new StringBuffer();
			for (int i = 0; i < v.length; i++)
			{
				buf.append(v[i]);
				if ((i + 1) < v.length)
					buf.append(",");
			}
			params.put(k, buf.toString());
		}
		return this;
	}

	public ParameterSet remove(ParameterDefinition k)
	{
		params.remove(k);
		return this;
	}
	
	public String getAppend() throws UnsupportedEncodingException
	{
		if (params.size() == 0)
			return "";

		StringBuffer ret = new StringBuffer();
		Iterator<ParameterDefinition> it = params.keySet().iterator();
		while (it.hasNext())
		{
			ParameterDefinition param = it.next();
			
			ret.append("&");
			ret.append(APIRequest.urlEncodePair(param.getName(), params.get(param)));
		}
		
		return ret.toString();
	}
	
}
