package com.untzuntz.ustackclientapi;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.UUID;

import com.untzuntz.ustackclientapi.APIRequest.Method;
import com.untzuntz.ustackclientapi.exceptions.APIException;
import com.untzuntz.ustackclientapi.exceptions.APIFileRetrieveException;
import com.untzuntz.ustackclientapi.exceptions.APIParameterValueException;
import com.untzuntz.ustackclientapi.exceptions.BadRequestException;
import com.untzuntz.ustackclientapi.exceptions.ClientAuthenticationException;
import com.untzuntz.ustackclientapi.exceptions.ClientAuthorizationException;
import com.untzuntz.ustackclientapi.exceptions.ClientSideException;
import com.untzuntz.ustackclientapi.exceptions.ServerSideException;
import com.untzuntz.ustackclientapi.exceptions.UserAuthenticationException;
import com.untzuntz.ustackclientapi.status.Status;

/**
 * A Java implementation of the UStack API
 * 
 * @author jdanner
 *
 */
public class UStack {

	public static String API_BASE;
	public static String CLIENT_VERSION;
	public static String APP_NAME;
	public static String API_VERSION;
	public static String apiKey;
	public static String clientId;

	/**
	 * Returns data from POST call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T post(String apiCall, Class<T> dataClass) throws ServerSideException, ClientSideException
	{
		return post(apiCall, dataClass, null);
	}
	
	/**
	 * Returns data from GET call
	 * 
	 * @param apiCall
	 * @param dataCollectionClass
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T get(String apiCall, Class<T> dataClass) throws ServerSideException, ClientSideException
	{
		return get(apiCall, dataClass, null);
	}
	/**
	 * Returns data from PUT call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T put(String apiCall, Class<T> dataClass) throws ServerSideException, ClientSideException
	{
		return put(apiCall, dataClass, null);
	}
	
	/**
	 * Returns data from DELETE call
	 * 
	 * @param apiCall
	 * @param dataCollectionClass
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T delete(String apiCall, Class<T> dataClass) throws ServerSideException, ClientSideException
	{
		return delete(apiCall, dataClass, null);
	}
	
	/**
	 * Returns data from POST call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T post(String apiCall, Class<T> dataClass, ParameterSet params) throws ServerSideException, ClientSideException
	{
		return post(apiCall, dataClass, params, null);
	}
	
	/**
	 * Returns data from GET call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T get(String apiCall, Class<T> dataClass, ParameterSet params) throws ServerSideException, ClientSideException
	{
		return get(apiCall, dataClass, params, null);
	}
	/**
	 * Returns data from PUT call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T put(String apiCall, Class<T> dataClass, ParameterSet params) throws ServerSideException, ClientSideException
	{
		return put(apiCall, dataClass, params, null);
	}
	
	/**
	 * Returns data from DELETE call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T delete(String apiCall, Class<T> dataClass, ParameterSet params) throws ServerSideException, ClientSideException
	{
		return delete(apiCall, dataClass, params, null);
	}
	
	/**
	 * Returns data from POST call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T post(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl) throws ServerSideException, ClientSideException
	{
		return post(apiCall, dataClass, params, apiBaseUrl, null, null);
	}
	
	/**
	 * Returns data from GET call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T get(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl) throws ServerSideException, ClientSideException
	{
		return get(apiCall, dataClass, params, apiBaseUrl, null, null);
	}
	/**
	 * Returns data from PUT call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @return
	 * @throws ServerSideException
	 */
	public static <T> T put(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl) throws ServerSideException, ClientSideException
	{
		return put(apiCall, dataClass, params, apiBaseUrl, null, null);
	}
	
	/**
	 * Returns data from DELETE call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @return
	 * @throws ServerSideException
	 * @throws ClientSideException 
	 */
	public static <T> T delete(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl) throws ServerSideException, ClientSideException
	{
		return delete(apiCall, dataClass, params, apiBaseUrl, null, null);
	}
	
	/**
	 * Returns data from POST call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @param clientId
	 * @param apiKey
	 * @return
	 * @throws ServerSideException
	 * @throws ClientSideException 
	 */
	public static <T> T post(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl, String clientId, String apiKey) throws ServerSideException, ClientSideException
	{
		return doHttpMethod(apiCall, dataClass, Method.POST, params, apiBaseUrl, clientId, apiKey);
	}
	
	/**
	 * Returns data from GET call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @param clientId
	 * @param apiKey
	 * @return
	 * @throws ServerSideException
	 * @throws ClientSideException 
	 */
	public static <T> T get(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl, String clientId, String apiKey) throws ServerSideException, ClientSideException
	{
		return doHttpMethod(apiCall, dataClass, Method.GET, params, apiBaseUrl, clientId, apiKey);
	}
	/**
	 * Returns data from PUT call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @param clientId
	 * @param apiKey
	 * @return
	 * @throws ServerSideException
	 * @throws ClientSideException 
	 */
	public static <T> T put(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl, String clientId, String apiKey) throws ServerSideException, ClientSideException
	{
		return doHttpMethod(apiCall, dataClass, Method.PUT, params, apiBaseUrl, clientId, apiKey);
	}
	
	/**
	 * Returns data from DELETE call
	 * 
	 * @param apiCall
	 * @param dataClass
	 * @param params
	 * @param apiBaseUrl
	 * @param clientId
	 * @param apiKey
	 * @return
	 * @throws ServerSideException
	 * @throws ClientSideException 
	 */
	public static <T> T delete(String apiCall, Class<T> dataClass, ParameterSet params, String apiBaseUrl, String clientId, String apiKey) throws ServerSideException, ClientSideException
	{
		return doHttpMethod(apiCall, dataClass, Method.DELETE, params, apiBaseUrl, clientId, apiKey);
	}

	/**
	 * Core Method for calling HTTP API
	 * @param apiCall
	 * @param dataClass
	 * @param method
	 * @return
	 * @throws ServerSideException
	 */
	@SuppressWarnings("unchecked")
	private static <T> T doHttpMethod(String apiCall, Class<T> dataClass, Method method, ParameterSet params, 
									String apiBase, String clientId, String apiKey) throws ClientSideException, ServerSideException
	{
		validateAPISettings(apiBase, clientId, apiKey);
		
		StringBuffer url = new StringBuffer();
		if (apiBase == null)
			url.append(API_BASE);
		else
			url.append(apiBase);
		
		if (API_VERSION != null)
			url.append(API_VERSION);
		
		url.append(apiCall);

		StringBuffer query = new StringBuffer();
		try {
			query.append(getAPIParams());
			if (params != null)
				query.append(params.getAppend());
		} catch (UnsupportedEncodingException uee) {
			throw new BadRequestException("Unable to encode parameters to " + APIRequest.CHARSET);
		}

		APIResponse resp = APIRequest.makeURLConnectionRequest(method, url.toString(), query.toString());

		if (resp.getConnection() != null)
		{
			if (dataClass.equals(InputStream.class))
				return (T)resp.getConnection();

			File tgtFile = new File(UUID.randomUUID().toString());
			OutputStream out = null;
			InputStream in = resp.getConnection();
			byte[] dataBuf = new byte[4096];
			try {
				
				out = new BufferedOutputStream(new FileOutputStream(tgtFile));
				int n = 0;
				while ((n = in.read(dataBuf)) != -1)
					out.write(dataBuf, 0, n);
				
				out.flush();
				
			} catch (IOException e) {
				throw new APIFileRetrieveException(e);
			} finally {
				try { out.close(); } catch (Exception ie) {}
				try { in.close(); } catch (Exception ie) {}
			}
			
			return (T)tgtFile;
		}
		else
		{
			int responseCode = resp.getCode();
			String responseBody = resp.getBody();
			verifyStatus(responseCode, responseBody);
			
			return APIRequest.gson.fromJson(responseBody, dataClass);
		}
	}
	
	/**
	 * Returns encoded client id and API key values pairs
	 * @return
	 * @throws UnsupportedEncodingException
	 */
	private static String getAPIParams() throws UnsupportedEncodingException
	{
		StringBuffer buf = new StringBuffer();
		buf.append(APIRequest.urlEncodePair("client_ver", CLIENT_VERSION)).append("&");
		buf.append(APIRequest.urlEncodePair("app_name", APP_NAME));
		
		if (clientId != null)
		{
			buf.append("&");
			buf.append(APIRequest.urlEncodePair("client_id", clientId)).append("&");
			buf.append(APIRequest.urlEncodePair("api_key", apiKey));
		}
		return buf.toString();
	}

	/**
	 * Check the status object. For anything not 'OK' throw an exception
	 * 
	 * @param responseCode
	 * @param responseBody
	 * @throws APIException
	 */
	private static void verifyStatus(int responseCode, String responseBody) throws APIException
	{
		//System.out.println(responseCode + " -> Error Response: " + responseBody);

		Status status = APIRequest.gson.fromJson(responseBody, Status.class);
		
		if (!"SUCCESS".equalsIgnoreCase(status.getApiResult().getStatus()))
		{
			if ("ParamValueException".equals(status.getType()))
			{
				ParamTypeExcep ret = APIRequest.gson.fromJson(responseBody, ParamTypeExcep.class);
				throw new APIParameterValueException(ret);
			}
			else if ("UserAuthenticationException".equals(status.getType()))
				throw new UserAuthenticationException(status.getMessage(), status.getFailReason());
			
			if ("Bad Client ID/API Key".equals(status.getApiResult().getMessage()))
				throw new ClientAuthenticationException(status.getApiResult().getMessage());
			
			if ("API Client Not Authorized".equals(status.getApiResult().getMessage()))
				throw new ClientAuthorizationException(status.getApiResult().getMessage());
			
			if ("Not Authorized".equals(status.getApiResult().getMessage()))
				throw new ClientAuthorizationException(status.getApiResult().getMessage());
			
			//System.err.println(responseCode + " - Status: " + status.getApiResult().getStatus() + " -> " + status.getApiResult().getMessage());
			throw new APIException(status.getApiResult().getMessage());
		}
	}

	public static class ParamTypeExcep {

		private String type;
		private String paramTypeName;
		private String message;
		
		public String getType() {
			return type;
		}
		public String getParamTypeName() {
			return paramTypeName;
		}
		public String getMessage() {
			return message;
		}

		
	}
	/**
	 * Check to verify the apiKey and clientId are set on this object
	 * 
	 * @throws ClientSideException
	 */
	private static void validateAPISettings(String apiBase, String paramClientId, String paramApiKey) throws ClientSideException
	{
		if ((API_BASE == null || API_BASE.length() == 0) && (apiBase == null || apiBase.length() == 0))
			throw new ClientSideException("No API_BASE provided. Set UStack.API_BASE on app startup.");

		if (CLIENT_VERSION == null || CLIENT_VERSION.length() == 0)
			throw new ClientSideException("No CLIENT_VERSION provided. Set UStack.CLIENT_VERSION on app startup.");

		if (APP_NAME == null || APP_NAME.length() == 0)
			throw new ClientSideException("No APP_NAME provided. Set UStack.APP_NAME on app startup.");

		if ((apiKey == null || apiKey.length() == 0 || clientId == null || clientId.length() == 0) &&
				(paramApiKey == null || paramApiKey.length() == 0 || (paramClientId == null || paramClientId.length() == 0)))
			throw new ClientSideException("No clientId/apiKey provided. Set UStack.clientId/UStack.apiKey on app startup.");
	}

	
}
